% The MIT License (MIT)
% Copyright (c) 2016 Elmar Rueckert, Daniel Tanneberg, David Kappel
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE. 

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert.elmar@gmail.com, tanneberg@ias.tu-darmstadt.de, david.kappel@igi.tugraz.at
% *
% **********************************************************

% % * Description: This demo shows how a simple recurrent neural network 
% %   can be used to generate spike trains without input. It implements 
% %   a one-dimensional dot movement that performs a random walk. 
% %   In the 2nd row idealized WTA sampling is implemented where only 1 out
% %   of K neurons can be active per time step (1ms). The 3rd row shows an
% %   approximation where multiple neurons can be simultaneously active. 

function RecurrentSamplingDemo()
global GLOBAL_OBJ_HANDLES

    % ##################################
    % network parameters 
    net.numNeurons = 30;                              % number of neurons 
    net.transferFunction = @(x) x;                    % simple identity function
    net.W = mk_stochastic( filter2( eye(net.numNeurons), ...
        fspecial('gaussian', net.numNeurons, 1.5) ) );% transition model
    net.b = zeros(net.numNeurons, 1);                 % no input
    net.init_state = round(net.numNeurons*rand(1,1)); % random initial state as natural number
    net.flgRunsFirstTime = 1;
    
    % ##################################
    % plotting parameters 
    net.numStepsVisualized = 100;
    net.pauseAfterPlottinInSec = 0.05;
    net.spikeTrain = zeros(net.numNeurons, net.numStepsVisualized);
    net.timeStepId = 1;
    net.spikeTrain(net.init_state,net.timeStepId) = 1;

    
    % ##################################
    % Generate a simple GUI to visualize the process. 
    disp('% * Hit [q] to terminate the program!');
    
    %*** Keystroke Event
    % Define the key press callback fexitunction and open a figure to detect key strokes
    GLOBAL_OBJ_HANDLES.flgSimIsRunning = 1;  
    GLOBAL_OBJ_HANDLES.hFigKeyStrokes = figure; 
    set(GLOBAL_OBJ_HANDLES.hFigKeyStrokes, 'color', 'white');
    set(GLOBAL_OBJ_HANDLES.hFigKeyStrokes,'KeyPressFcn',@KeyPressFcn);

    % ##################################
    % STATIC PLOTS
    mySpikeMap = gray(2);
    mySpikeMap = mySpikeMap(end:-1:1,:);
    myWeightMap = jet(20);
    
    figure(GLOBAL_OBJ_HANDLES.hFigKeyStrokes);
    figurePosition = get(GLOBAL_OBJ_HANDLES.hFigKeyStrokes, 'Position');
    
    GLOBAL_OBJ_HANDLES.hFigPanelLeftSmall = subplot(3,3,1);
    imagesc(net.W);
    %plotModel(taskLearn.W,  GLOBAL_OBJ_HANDLES.hFigPanelLeftSmall);
    
    colormap(GLOBAL_OBJ_HANDLES.hFigPanelLeftSmall, myWeightMap);
    set(GLOBAL_OBJ_HANDLES.hFigPanelLeftSmall,'xtick',[],'ytick',[], 'box', 'off'); axis off;
    GLOBAL_OBJ_HANDLES.lTextBox = uicontrol('style','text');
    set(GLOBAL_OBJ_HANDLES.lTextBox,'String','transition model');
    set(GLOBAL_OBJ_HANDLES.lTextBox, 'Position', [60,390,160,20],'BackgroundColor','white');
    
    GLOBAL_OBJ_HANDLES.hFigPanelMidSmall = subplot(3,3,2);axis off; 
    set(GLOBAL_OBJ_HANDLES.hFigPanelMidSmall,'xtick',[],'ytick',[])
    
    GLOBAL_OBJ_HANDLES.hFigPanelRightSmall = subplot(3,3,3); axis off;
    set(GLOBAL_OBJ_HANDLES.hFigPanelRightSmall,'xtick',[],'ytick',[])
    GLOBAL_OBJ_HANDLES.rTextBox = uicontrol('style','text');
    set(GLOBAL_OBJ_HANDLES.rTextBox,'String','membrane pot.');
    set(GLOBAL_OBJ_HANDLES.rTextBox, 'Position', [370,390,160,20],'BackgroundColor','white');
    
    
    GLOBAL_OBJ_HANDLES.hFigPanelBottom = subplot(3,3,[4,5,6]);
    set(GLOBAL_OBJ_HANDLES.hFigPanelBottom,'xtick',[],'ytick',[])
    subplot(GLOBAL_OBJ_HANDLES.hFigPanelBottom); imagesc(net.spikeTrain);
    colormap(GLOBAL_OBJ_HANDLES.hFigPanelBottom, mySpikeMap);
    GLOBAL_OBJ_HANDLES.RowTwoTextBox = uicontrol('style','text');
    set(GLOBAL_OBJ_HANDLES.RowTwoTextBox,'String','ideal WTA model where a single neuron is active per time step');
    set(GLOBAL_OBJ_HANDLES.RowTwoTextBox, 'Position', [40,270,430,20],'BackgroundColor','white', ...
        'HorizontalAlignment','left');
    
    GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree = subplot(3,3,[7,8,9]);
    set(GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree,'xtick',[],'ytick',[])
    subplot(GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree); imagesc(net.spikeTrain);
    colormap(GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree, mySpikeMap);
    GLOBAL_OBJ_HANDLES.RowThreeTextBox = uicontrol('style','text');
    set(GLOBAL_OBJ_HANDLES.RowThreeTextBox,'String','approximated model where multiple neurons may be active per time step');
    set(GLOBAL_OBJ_HANDLES.RowThreeTextBox, 'Position', [40,140,430,20],'BackgroundColor','white', ...
        'HorizontalAlignment','left');
    
    %TEMPERATURE CONTROL
    GLOBAL_OBJ_HANDLES.beta = 20;
    b = uicontrol('Parent',GLOBAL_OBJ_HANDLES.hFigKeyStrokes,'Style','slider','Position',...
        [260,300,20,90],...q
              'value', GLOBAL_OBJ_HANDLES.beta, 'min',1, 'max',100, ...
              'Callback', @temperatureCallback);
    GLOBAL_OBJ_HANDLES.mTextBox = uicontrol('style','text');
    set(GLOBAL_OBJ_HANDLES.mTextBox,'String',['temp. = ' num2str(GLOBAL_OBJ_HANDLES.beta)]);
    set(GLOBAL_OBJ_HANDLES.mTextBox, 'Position', [220,390,100,20],'BackgroundColor','white');
            
    
    netTheory = net;
    netApprox = net;
    
    while GLOBAL_OBJ_HANDLES.flgSimIsRunning

        %select the sampling method
        netTheory = runWTASampling(netTheory);
        netApprox = runStdSampling(netApprox);

        pause(net.pauseAfterPlottinInSec);
    end
end


function net = runStdSampling(net)
%This function implements sampling recurrent neural networks with poisson point neurons. 
global GLOBAL_OBJ_HANDLES

    if net.flgRunsFirstTime
        net.flgRunsFirstTime = 0;
        net.b = 2*ones(size(net.b));
        net.W = -2+net.W;
    end

    % ##################################
    % sample from the network and update the plot
%     while GLOBAL_OBJ_HANDLES.flgSimIsRunning
        net.timeStepId = net.timeStepId + 1;
        if net.timeStepId > net.numStepsVisualized
            net.timeStepId = 2;
            tmp = net.spikeTrain(:,end);
            net.spikeTrain = zeros(net.numNeurons, net.numStepsVisualized);
            net.spikeTrain(:,1) = tmp;
        end
        
        %normalized exponential transfer function
        net.transferFunction = @(u) exp(1.50*GLOBAL_OBJ_HANDLES.beta*u)./sum(exp(1.50*GLOBAL_OBJ_HANDLES.beta*u)); %normalized exp
        
        net.current_spike_vector = net.spikeTrain(:,net.timeStepId-1);
        sampleSize = 1; 
        [net.spikeTrain(:,net.timeStepId), u] = drawSample(net, net.current_spike_vector, sampleSize);
         
%         stem(GLOBAL_OBJ_HANDLES.hFigPanelRightSmall, u);
%         axis(GLOBAL_OBJ_HANDLES.hFigPanelRightSmall,[1,net.numNeurons,0,1]);

        if sum(sum(net.spikeTrain)) > 0
            subplot(GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree); imagesc(net.spikeTrain);
        else
            ;
        end

        set(GLOBAL_OBJ_HANDLES.hFigPanelBottomRowThree,'xtick',[],'ytick',[], 'box', 'off'); axis off;
%         pause(net.pauseAfterPlottinInSec);
%     end
end

function net = runWTASampling(net)
%This function implements neural sampling in wta circuits. A neuron stays active for a refractory period. 
global GLOBAL_OBJ_HANDLES

    if net.flgRunsFirstTime
        net.flgRunsFirstTime = 0;
        net.refractoryPeriodeInSeconds = 12e-3;%12e-3; %10ms
        net.simTimeStep = 1e-3;
        tau = net.refractoryPeriodeInSeconds/net.simTimeStep;

        net.refractoryState = zeros(net.numNeurons,1);
        net.refractoryState(net.init_state) = tau;
        net.current_state = net.init_state;
    end
    % ##################################
    % sample from the network and update the plot
%     while GLOBAL_OBJ_HANDLES.flgSimIsRunning

        net.timeStepId = net.timeStepId + 1;
        if net.timeStepId > net.numStepsVisualized
            net.timeStepId = 2;
            net.spikeTrain = zeros(net.numNeurons, net.numStepsVisualized);
            if~isempty(net.current_state)
                net.spikeTrain(net.current_state,net.timeStepId-1) = 1;
            end
        end
        
        active = find(net.refractoryState > 0);
        inactive = find(net.refractoryState <= 0);
        
        %advance internal state of neurons that are in their ref. periode
        net.refractoryState(active) = max(zeros(length(active),1),-1+net.refractoryState(active));
        net.current_spike_vector = net.spikeTrain(:,net.timeStepId-1);%net.refractoryState > 0;
        
        %update membran potential of all inactive neurons
        u = net.W(:,inactive)'*net.current_spike_vector + net.b(inactive);
        
        %WTA 
        if sum(active) == 0
           %u=u-max(u);
           u=exp(GLOBAL_OBJ_HANDLES.beta*u);
           if sum(isnan(u)) > 0
               error('bad temperature values');
           end
           u=u./sum(u);
           dr=histc(rand(1,1),[0; cumsum(u)]);
           net.current_spike_vector = dr(1,1:end-1)';
           if sum(net.current_spike_vector) < 1
                keyboard
           end
        else
           net.current_spike_vector = zeros(net.numNeurons,1);
        end 
        net.spikeTrain(:,net.timeStepId) = net.current_spike_vector;
        net.current_state = find(net.current_spike_vector); 
        
        stem(GLOBAL_OBJ_HANDLES.hFigPanelRightSmall, u);
        axis(GLOBAL_OBJ_HANDLES.hFigPanelRightSmall,[1,net.numNeurons,0,1]);
        
        subplot(GLOBAL_OBJ_HANDLES.hFigPanelBottom); imagesc(net.spikeTrain);
        if ~isempty(net.current_state)
            hold on;
            plot([1;net.timeStepId],[net.current_state;net.current_state],'r', 'linewidth',3);
            hold off;
        end
        
        set(GLOBAL_OBJ_HANDLES.hFigPanelBottom,'xtick',[],'ytick',[], 'box', 'off'); axis off;
%         pause(net.pauseAfterPlottinInSec);
%     end
end


function KeyPressFcn(hObject, eventdata, handles)
global GLOBAL_OBJ_HANDLES

    pressedKeyChar = eventdata.Character;

    if pressedKeyChar == 'q' %terminate the simulation
        GLOBAL_OBJ_HANDLES.flgSimIsRunning = 0; 
        close(GLOBAL_OBJ_HANDLES.hFigKeyStrokes);
   elseif pressedKeyChar == 'i' 
         %not implemented yet
   end
end

function temperatureCallback(src,evt)
global GLOBAL_OBJ_HANDLES
   %src.Color = 'red';
   GLOBAL_OBJ_HANDLES.beta = round(src.Value);
   set(GLOBAL_OBJ_HANDLES.mTextBox,'String',['temp. = ' num2str(GLOBAL_OBJ_HANDLES.beta)]);
end
