% The MIT License (MIT)
% Copyright (c) 2016  Daniel Tanneberg, Elmar Rueckert, David Kappel
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE. 

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: tanneberg@ias.tu-darmstadt.de, rueckert.elmar@gmail.com, david.kappel@igi.tugraz.at
% *
% **********************************************************

function ImitationLearningDemo()


% load 2DTrainSamples 
trainingData = load('trainData_KUKA_2Dmovements_spikeTrains');
trainingData = trainingData.trainData;
rng(5); 

% load learned model using all training data for comparison
fullModel = load('learnedModel_2D_imitation.mat');
fullModel = fullModel.taskLearn.W;

% ##################################
% parameters 
taskLearn = {};
taskLearn.numNeurons = 225; % needs to map the training data
taskLearn.transferFunction = @(x) x;
refractoryPeriode = 10; %ms 

rate = 0.1; % learningRate
samples_per_it = 1; % number of samples used for model expectation

show_progress = 1;
visualizeLearningSlow = show_progress && 1;
visualizeEnd = 1;
visualizeIntervall = 1000; % showed ms, whenever visualizeSlow is shown and steps for calculating the error


% init W with strong random inhibition
taskLearn.W = sparse(taskLearn.numNeurons, taskLearn.numNeurons);
selfConnectionMatrix = ones(taskLearn.numNeurons, taskLearn.numNeurons) - eye(taskLearn.numNeurons, taskLearn.numNeurons);
a = -0.99;
b = -0.95;
r = (b-a).*rand(taskLearn.numNeurons, taskLearn.numNeurons) + a;
taskLearn.W = r .* selfConnectionMatrix;

taskLearn.errors = [];
[error, mse, rmse] = learnError(fullModel , taskLearn.W);
taskLearn.errors = [taskLearn.errors; error, mse, rmse];

% init b
initB = -0.7;
taskLearn.b = repmat(initB,taskLearn.numNeurons, 1);

% prepare training data
trainSpikeTrains = {};
fields = fieldnames(trainingData.data);

for i=1:numel(fields)
    spikes = trainingData.data.(fields{i}).spikeTrain{1};
    % ##### only 7s subset for each trial for fast testing, use complete 'spike' to use all data
    trainSpikeTrains{i} = spikes(:,1:5000); 
end
numTrainData = numel(fields); 

% ##################################
 
if visualizeLearningSlow
    figure;
    imagesc(taskLearn.W);
    cMap = cbrewer('div','RdYlBu',9); 
    cMap = cMap(end:-1:1,:);
    caxis([-1 1]);
    colormap(cMap);
    colorbar;
    title(sprintf('Time: %ds',0));
    drawnow();
end

%% learning progress
% ### show progess
dateFormat = 'HH:MM:SS';
date_str = datestr(datetime(), dateFormat);
fprintf('Start time: %s\n',date_str);
fprintf('#TrainData: %d | Samples for Exp.: %d\n',numTrainData, samples_per_it);

tic

% ###  number of training data trajectories
for trainIndex=1:numTrainData   
    oneTraj = trainSpikeTrains{trainIndex};
    
    if show_progress
        fprintf('%d/%d (%.1fs) Progress: 0%%',trainIndex,numTrainData,size(oneTraj,2)/1000);
        prg = 0;
    end
      
    % ### iterate over the given movement, replay
    for t=refractoryPeriode+1:size(oneTraj,2)
        
        % contrastive divergence
        
        % ### create trainData vector of pre and postsynaptic spikes
        preSynSpikes = sum(oneTraj(:,t-refractoryPeriode:t-1),2);
        postSynSpikes = oneTraj(:,t);
        trainData = [preSynSpikes, postSynSpikes];

        % ### calculate the data expectation
        trainExp = trainData(:,1)*trainData(:,2)';
  
        % ### draw sample with model
        learnSample = drawSample(taskLearn, trainData(:,1), samples_per_it);
        modelExp = 0;
        for s=1:samples_per_it
            % ### calc the model expectation
            modelExp = modelExp + preSynSpikes*learnSample(:,s)';
        end

        % ### calculate gradient and update model
        grad = trainExp - (modelExp/samples_per_it);
        taskLearn.W = (taskLearn.W + rate*grad) .* selfConnectionMatrix;
        
        
        if visualizeLearningSlow && mod(t,visualizeIntervall) == 0;
            [error, mse, rmse] = learnError(fullModel , taskLearn.W);
            taskLearn.errors = [taskLearn.errors; error, mse, rmse];
            
            % ### visualize learning
            imagesc(taskLearn.W);
            cMap = cbrewer('div','RdYlBu',9); 
            cMap = cMap(end:-1:1,:);
            caxis([-1 1]);
            colormap(cMap);
            colorbar;
            title(sprintf('TrainData: %d/%d | Time: %d/%ds',trainIndex,numTrainData,t/1000,size(oneTraj,2)/1000));
            drawnow();
        end
              
        if show_progress 
            new_prg = floor(100*t / size(oneTraj,2));
            if prg ~= new_prg && mod(new_prg,5) == 0 
                if prg == 0
                    prg = 1;
                end
                fprintf(repmat('\b',1,floor(log10(prg))+2));
                fprintf('%d%%',new_prg);
                prg = new_prg;
            end
        end
    end
    fprintf('\n');
end

date_str = datestr(datetime(), dateFormat);
fprintf('End time: %s\n',date_str);
fprintf('Duration: %fs \n',toc);


% ### show result
if visualizeEnd
    if visualizeLearningSlow       
        close;
    end
    %%
    hfig = figure;
    imagesc(taskLearn.W);
    cMap = cbrewer('div','RdYlBu',9); 
    cMap = cMap(end:-1:1,:);
    caxis([-1 1]);
    colormap(cMap);
    colorbar;
    title(sprintf('TrainData: %d (samples for exp.: %d) learnRate: %.2f ',numTrainData,samples_per_it, rate));
 
    
    figure; hold on;  
    line([0 size(taskLearn.errors,1)], [taskLearn.errors(end,3) taskLearn.errors(end,3)],'Color','r');
    plot(taskLearn.errors(:,3)); 
    xlabel(sprintf('Replay time (x%ds)',visualizeIntervall/1000));
    ylabel('error');
    title(sprintf('RMSE (last: %.3f)',taskLearn.errors(end,3)));
end

end

