% The MIT License (MIT)
% Copyright (c) 2016 David Kappel, Elmar Rueckert, Daniel Tanneberg 
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE. 

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: david.kappel@igi.tugraz.at, rueckert.elmar@gmail.com, tanneberg@ias.tu-darmstadt.de
% *
% **********************************************************

% % * Description: This demo shows how a simple recurrent neural network 
% %   can be used to solve the finite horizon planning problem illustrated
% in Figure 1C in the paper. 

function ReinforcementLearningTVInputDemo()

N = 30;            % number of time steps
K = 15;            % number of state neurons
L = K+1;           % number of observations

transmat = mk_stochastic( filter2( eye(K), fspecial('gaussian', K, 1.5) ) );
obsmat = [ eye(K), (1/K)*ones(K,1) ];

init_state = (1/K)*ones(1,K);

%init_state = zeros(1,K);
%init_state(1) = 1;

figure; 
subplot(1,2,1); imagesc(transmat); title('transition matrix');
subplot(1,2,2); imagesc(obsmat); title('observation matrix');


obs_desired = [L*ones(1,N/2-1), 2, L*ones(1,N/2-1), K-1]; 
obslik = multinomial_prob(obs_desired, obsmat);
[alpha, beta, gamma, loglik, xi_summed, gamma2] = fwdback(init_state, transmat, obslik);

beta = mk_stochastic((beta+eps)')';
gamma = mk_stochastic((gamma+eps)')';

figure;
subplot(3,1,1); imagesc( alpha ); title('forward messages optimal');
subplot(3,1,2); imagesc( beta ); title('backward messages optimal');
subplot(3,1,3); imagesc( gamma ); title('likelihood optimal');
drawnow 

V = log(mk_stochastic(0.8+0.01*randn(K,N)')');
V_z = zeros(K,N);

num_trials = 10000;

h = figure;

kl_div_beta = nan(1,num_trials);
kl_div_gamma = nan(1,num_trials);
success_trials = zeros(1,num_trials);

beta_est = mk_stochastic( exp(V)' )';

max_learning_rate = 0.05;

s_rate = 0;

for trial = 1:num_trials

    % forward sampling
    state = zeros(1,N);
    state(1)  = wta_draw(init_state);
    obs = zeros(1,N);
    for n=2:N
        state(n) = wta_draw(normalise( transmat(state(n-1),:).*exp(V(:,n))' ));
        %state(n) = wta_draw( normalise( transmat(state(n-1),:) ) );
        obs(n) = wta_draw(obsmat(state(n),:));
    end

    cons = ones(1,N);
    cons(obs_desired<L) = obs(obs_desired<L) == obs_desired(obs_desired<L);
    %con_ok_id = max(1,N - find( cumprod( [cons(end:-1:2),0] )==0, 1 ));
    
    if all(cons) % hard constraints
        success_trials(trial) = 1;
        spike_train = repmat(state,K,1)==repmat((1:K)',1,N);
        eta = min(1/(sum(success_trials)^(2/3)),max_learning_rate);        
        V = V + eta*( exp(-V).*spike_train - 1/K );
        % beta_est = mk_stochastic( exp(V)' )';
        % problem with normalization... lets see...
        beta_est = exp(V);
    end
    
    %kl_div_beta(trial) = sum(sum(beta_est.*(log(beta_est)-log(beta))));
    kl_div_beta(trial) = sum(sum(beta.*(log(beta)-log(beta_est))));
    kl_div_gamma(trial) = sum(sum(gamma.*(log(gamma)-log(beta_est))));
    
    if trial>100
        s_rate = sum(success_trials(trial-99:trial));
    end

    if mod(trial,100) == 0
        figure(h);
        subplot(2,1,1); plot(state); ylim([1,K]); xlim([1,N]); set(gca,'YDir','reverse');
        ylabel('state');  title(sprintf('iteration %i', trial));
        subplot(2,1,2); imagesc(beta_est);
        xlabel('time'); title(sprintf('num seccuess trials: %i   success rate (last 100): %i/100', sum(success_trials), s_rate));
        drawnow
    end
end

%figure; plot(kl_div_beta); xlabel('iteration'); ylabel('Kullback-Leibler Divergence (unnormalized)');
figure; plot(cumsum(success_trials)); xlabel('iteration'); ylabel('number of successes');

figure;
num_test_trials = 50;
for trial = 1:num_test_trials

    % forward sampling
    state = zeros(1,N);
    state(1)  = wta_draw(init_state);
    obs = zeros(1,N);
    for n=2:N
        state(n) = wta_draw(normalise( transmat(state(n-1),:).*exp(V(:,n))' ));
        %state(n) = wta_draw( normalise( transmat(state(n-1),:) ) );
        obs(n) = wta_draw(obsmat(state(n),:));
    end
    plot(state); hold on; ylim([1,K]); xlim([1,N]); set(gca,'YDir','reverse');
    pause(0.1);
    drawnow
end

end
