% The MIT License (MIT)
% Copyright (c) 2016 Elmar Rueckert, Daniel Tanneberg, David Kappel
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE.

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert.elmar@gmail.com, tanneberg@ias.tu-darmstadt.de, david.kappel@igi.tugraz.at
% *
% **********************************************************

function RealRobotReachingDemo()

flgUseTheParallelToolboxForSpeedUp = 1; 

% load model #############################################################
% load and set the learned transition model
taskModel = load('learnedModel_2D_imitation'); 
taskModel = taskModel.taskLearn.W;


rng(5); % For Debugging only
% parameters #############################################################

% parameters for rejecting sampling and shortening
delta = 0.2; % precision threshold
time_window = 20; % time window to consider for precision threshold
threshold = 0.15; % jerk threshold
cut_window = 20; % time window to consider for shortening of the trajectories

% model and task parameters
parameters = {};
parameters.taskModel = taskModel;
parameters.numStateNeuronsPerDim = sqrt(size(taskModel,1)); % number of state neurons per dimension
parameters.stateDimensions = 2;
% timing parameters, how long initial position neurons are active and total
% sample length
parameters.numStepsInitialStateOn = 300; % ms
parameters.numTimeSteps = 1000; % ms 
         
parameters.bAddConstant = -0.9; % additive constant for the membrane potential, potential offset 
parameters.additiveConstantStateW = 0.1; % additive constant on the synaptic weights of the state neurons
parameters.additiveConstantTaskW = -0.3; % additive constant on the synaptic weights of the task neurons

% membrane potential offsets for task neurons
parameters.bInitialStateOn = 0.05 - parameters.bAddConstant; % initial position active
parameters.bInitialStateOff = -10; % initial position inactive
parameters.bTargetStateOn = 0.05 - parameters.bAddConstant; % target position active

% scaling factor for determining the synaptic weights from task to state
% neurons based on Euclidean distance
parameters.distScalingFactorTask = 0.75; 

% initital and target position           
parameters.initialState = [-0.53; 0.42];
parameters.targetState = [0.64; -0.71];

% flag for obstacles
parameters.isObstacle = 1;
% create static obstacles, indicated by the IDs of the covered state
% neurons, static obstacles are active during the whole sample length
% (obstacle_times = -10 for all timesteps)
parameters.obstacle_ids = [126:129, 111:113, 88, 89, 103, 104]'; 
parameters.obstacle_times = parameters.bAddConstant * ones(size(parameters.obstacle_ids,1), parameters.numTimeSteps); 
parameters.obstacle_times(1:numel(parameters.obstacle_ids),:) = -10;

task = TwoDimTask(parameters);

% create net #############################################################
net = GenerateSNNDiscreteTask(task);

% sample


% number of sampled movements
samples = 50;

spikeTrains = {};
container = {};
movementPlansTemp = {};

if flgUseTheParallelToolboxForSpeedUp
    parfor ss=1:samples
        fprintf('Sample %d/%d \n',ss,samples);
        % draw samples by simulating the network dynamics
        spikeTrains{ss} = net.simulate(task.numTimeSteps);
        % transform spiketrains into movement trajectories
        movementPlansTemp{ss} = task.MapSpikesToTaskFeatures(SNNGaussianFilter(spikeTrains{ss})); 
    end
else
    for ss=1:samples
        fprintf('Sample %d/%d \n',ss,samples);
        % draw samples by simulating the network dynamics
        spikeTrains{ss} = net.simulate(task.numTimeSteps); % spikeTrains cell structure
        % transform spiketrains into movement trajectories
        movementPlansTemp{ss} = task.MapSpikesToTaskFeatures(SNNGaussianFilter(spikeTrains{ss}));         
    end
end

container.spikeTrains = spikeTrains;
container.movementPlans = movementPlansTemp; 

% reject samples
container.MPs = [];
container.accepted = [];
container.accepted_MPs = [];
container.accepted_cuts = [];

container.rejected = [];
container.rejected_MPs = [];


for s=1:samples
    
    % rejecting samples
    [~, ~, ids] = rejectSamples([container.movementPlans{s}(1,:);container.movementPlans{s}(2,:)], task, delta, time_window, threshold);  

    if sum(ids) == 1
        container.accepted = [container.accepted,s];
        container.accepted_MPs = [container.accepted_MPs;container.movementPlans{s}(1,:);container.movementPlans{s}(2,:)];
        % shorten samples, i.e., cutting of unprecise start and end part
        [s_start, s_end] = cutTraj([container.movementPlans{s}(1,:);container.movementPlans{s}(2,:)], task,cut_window);
        container.accepted_cuts = [container.accepted_cuts; s_start, s_end];
    else
        container.rejected = [container.rejected,s];
        container.rejected_MPs = [container.rejected_MPs;container.movementPlans{s}(1,:);container.movementPlans{s}(2,:)]; 
    end
    container.MPs = [container.MPs; container.movementPlans{s}(1,:); container.movementPlans{s}(2,:)];
end

task.accepted = container.accepted;

    
%%  plotting
numAccepted = size(container.accepted_cuts,1);
endEffectorTraj = container.accepted_MPs;

hfig = task.PlotMovementPlan(container.accepted_MPs,container.accepted_cuts,'accepted');
set(gca,'ydir','reverse')

if 1 %plot rejected routes
    % set cut positions to the total sample length for the rejected samples for plotting
    cuts = ones(numel(container.rejected),2);
    cuts(:,2) = 1000;
    hfig = task.PlotMovementPlan(container.rejected_MPs,cuts,'rejected');
    set(gca,'ydir','reverse')
end

disp('% **********************************************************');
disp('% * ');
disp(['% * Mental Planning Process Completed, FOUND ' num2str(numAccepted) ' Routes to the Target!']);
disp('% * ');
fprintf('Press any key to start the robot arm simulation...\n');
pause

% * Prepare Robot Visualization
% **********************************************************
hndArmVisualizer = plotTwoLinkRobotArm();
%add the obstacles
figure(hndArmVisualizer.figureHnd);
%initial and target state
hold on;
scatter(task.initialState(1),task.initialState(2),400,'ko','LineWidth',5);
hold off;
hold on;
scatter(task.targetState(1),task.targetState(2),400,'kd','LineWidth',5);
hold off;
% obstacles
if task.isObstacle == 1
    for i=1:numel(task.obstacle_ids)
        id = task.obstacle_ids(i);
        hold on;
        scatter(task.neuronIdToStateMapping(1,id),task.neuronIdToStateMapping(2,id),800,[0.7 0.7 0.7],'s','filled');
        hold off;
    end
end

% * Plot Routes to the target
% **********************************************************
for trajId = 1:numAccepted
    stFig = plotSpikeTrain(container.spikeTrains{container.accepted(trajId)});
    hndArmVisualizer = plotTwoLinkRobotArm(endEffectorTraj(2*(trajId-1)+1:2*(trajId-1)+2,:), hndArmVisualizer, 1);   
    
    if trajId < numAccepted
        if input(sprintf('Do you want to continue with the next trajectory? [y/n]'),'s') == 'n'
            return;
        end
        close(stFig);
    end

end

