% The MIT License (MIT)
% Copyright (c) 2016 Elmar Rueckert, Daniel Tanneberg, David Kappel
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE.

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert.elmar@gmail.com, tanneberg@ias.tu-darmstadt.de, david.kappel@igi.tugraz.at
% *
% **********************************************************

function task = TwoDimTask(parameters)
if nargin < 1
    parameters = {};
end
task = parseParametersTwoDim(parameters);


% CARTESIAN MODEL #########################################################
task.numStateNeurons = task.numStateNeuronsPerDim^task.stateDimensions;
task.numNeurons = task.numStateNeurons + task.numTaskNeuronsPerConstraint*task.numConstraints;

% cartesian model
rangesPerDim = reshape(task.stateSpaceLimits,2,task.stateDimensions)';

% for two dimensions, align state neurons in grid
tmp = linspaceNDim(rangesPerDim(:,1), rangesPerDim(:,2), task.numStateNeuronsPerDim);
[X1,X2] = ndgrid(tmp(1,:), tmp(2,:));         
task.neuronIdToStateMapping = [X1(:)'; X2(:)'];

%ids
task.StateNeuronIds = 1:task.numStateNeurons;
task.InitialStateNeuronIds = task.numStateNeurons+1:task.numStateNeurons + task.numTaskNeuronsPerConstraint;
task.TargetStateNeuronIds = task.numTaskNeuronsPerConstraint + task.InitialStateNeuronIds;

task.cartW(1:task.numStateNeurons, 1:task.numStateNeurons) = task.taskModel;

%Task Neurons
x0s = repmat(task.initialState,1,task.numTaskNeuronsPerConstraint) ...
    + 0.01*randn(task.stateDimensions,task.numTaskNeuronsPerConstraint); % initial position
xTs = repmat(task.targetState,1,task.numTaskNeuronsPerConstraint) ...
    + 0.01*randn(task.stateDimensions,task.numTaskNeuronsPerConstraint); % target position

task.InitNeurons = x0s;
task.TargetNeurons = xTs;

targetIds = task.StateNeuronIds;

for i=task.InitialStateNeuronIds
    w_ji = computeDistancesInStateSpace(x0s(:,i-task.InitialStateNeuronIds(1)+1), ...
       task.neuronIdToStateMapping);
    
    valid = ~isnan(w_ji);
    task.cartW(i,targetIds(valid))   = w_ji(valid);   
end
for i=task.TargetStateNeuronIds
    w_ji = computeDistancesInStateSpace(xTs(:,i-task.TargetStateNeuronIds(1)+1), ...
       task.neuronIdToStateMapping);
    valid = ~isnan(w_ji);
    task.cartW(i,targetIds(valid))   = w_ji(valid);   
end

% transform distance into weights
WInit = task.cartW(task.InitialStateNeuronIds,1:task.numStateNeurons);
WInit = WInit * task.distScalingFactorTask / max(WInit(:));
WInit = bsxfun(@minus,ones(task.numTaskNeuronsPerConstraint,task.numStateNeurons) , WInit);
task.cartW(task.InitialStateNeuronIds,1:task.numStateNeurons) = WInit;

WTar = task.cartW(task.TargetStateNeuronIds,1:task.numStateNeurons);
WTar = WTar * task.distScalingFactorTask / max(WTar(:));
WTar = bsxfun(@minus,ones(task.numTaskNeuronsPerConstraint,task.numStateNeurons) , WTar);
task.cartW(task.TargetStateNeuronIds,1:task.numStateNeurons) = WTar;

% state neurons additive W
WTrans = task.cartW(task.StateNeuronIds,task.StateNeuronIds);
WTrans(WTrans~=0) = WTrans(WTrans~=0) + task.additiveConstantStateW;
task.cartW(task.StateNeuronIds,task.StateNeuronIds) = WTrans;

% task neurons additive W
WTask = task.cartW(task.numStateNeurons+1:task.numNeurons,1:task.numStateNeurons);
WTask = WTask + task.additiveConstantTaskW;
WTask = max(0,WTask);
task.cartW(task.numStateNeurons+1:task.numNeurons,1:task.numStateNeurons) = WTask;

% add zeros to make W square again
task.cartW = [task.cartW,zeros(size(task.cartW,1), task.numNeurons-task.numStateNeurons)];

% create b, offset
task.cartb = task.bAddConstant*ones(task.numNeurons, task.numTimeSteps);
task.cartb(task.InitialStateNeuronIds,1:task.numStepsInitialStateOn) = task.cartb(task.InitialStateNeuronIds,1:task.numStepsInitialStateOn) + task.bInitialStateOn;
task.cartb(task.InitialStateNeuronIds,task.numStepsInitialStateOn+1:end) = task.cartb(task.InitialStateNeuronIds,task.numStepsInitialStateOn+1:end) + task.bInitialStateOff;
task.cartb(task.TargetStateNeuronIds,1:task.numStepsInitialStateOn) = task.cartb(task.TargetStateNeuronIds,1:task.numStepsInitialStateOn)  + task.bInitialStateOff;
task.cartb(task.TargetStateNeuronIds,task.numStepsInitialStateOn+1:end) = task.cartb(task.TargetStateNeuronIds,task.numStepsInitialStateOn+1:end) + task.bTargetStateOn;

task.cartW = max(-1,task.cartW);
task.cartb = max(-10,task.cartb);

if task.isObstacle == 1  
    for obs=1:numel(task.obstacle_ids)
        task.cartb(task.obstacle_ids(obs),:) = task.obstacle_times(obs,:);
    end
end


task.PlotMovementPlan = @(movementPlan_,cuts_,title_)localPlotMovementPlan(movementPlan_, cuts_, task,title_);
task.MapSpikesToTaskFeatures = @(spikeTrain_)localMapSpikesToTaskFeatures(spikeTrain_, task);
end

% compute distances between given positions
function w = computeDistancesInStateSpace(xOrigin, xTargets)
    w = pdist2(xOrigin',xTargets');
end

% decode spiketrain into spatial locations
function movementPlan=localMapSpikesToTaskFeatures(spikeTrain, task)
   
    spikeTrain = spikeTrain(task.StateNeuronIds,:);

    %Weighted Average
    movementPlan = nan(task.stateDimensions, size(spikeTrain,2));
    for t=1:size(spikeTrain,2)
        winnerIds = find(spikeTrain(:,t)>0);
        xs = task.neuronIdToStateMapping(:, winnerIds);
        if isempty(winnerIds)
            break;
        end
        if length(winnerIds) > 1
            w = spikeTrain(winnerIds,t)';
            w = w./sum(w);
            movementPlan(:,t) = sum(repmat(w,task.stateDimensions,1).*xs,2);
        else
            movementPlan(:,t) = xs;
        end
    end   
end



function hfig = localPlotMovementPlan(movementPlan, cuts, task, title_)

    hfig=figure;
    hold all;
    set(hfig, 'color','white');
    set(hfig,'PaperPositionMode','auto');
    set(gca, 'FontSize' , 20);

    % inital and target state
    scatter(task.initialState(1),task.initialState(2),400,'ko','LineWidth',5);
    scatter(task.targetState(1),task.targetState(2),400,'kd','LineWidth',5);
    
    % state neurons
    plot(task.neuronIdToStateMapping(1,:)', task.neuronIdToStateMapping(2,:)', '.k');
    for i=1:size(task.neuronIdToStateMapping,2)
        text(task.neuronIdToStateMapping(1,i),task.neuronIdToStateMapping(2,i), num2str(i), 'FontSize',6);
    end
    
    % task neurons
    plot(task.InitNeurons(1,:), task.InitNeurons(2,:), '.k');
    plot(task.TargetNeurons(1,:), task.TargetNeurons(2,:), '.k');

    % obstacles
    if task.isObstacle == 1
        for i=1:numel(task.obstacle_ids)
            id = task.obstacle_ids(i);
            scatter(task.neuronIdToStateMapping(1,id),task.neuronIdToStateMapping(2,id),780,[0.7 0.7 0.7],'s','filled');
        end
    end
    
    % movement plans
    mps = size(movementPlan,1)/2;
    x = [];
    y = [];
    for i=1:mps
       t = (i*2)-1;
       x = movementPlan(t,:);
       y = movementPlan(t+1,:);

       plot(x(cuts(i,1):cuts(i,2)), y(cuts(i,1):cuts(i,2)),'.-');
    end 
           
    
    axis(1.1*task.stateSpaceLimits);
    xlabel('x');
    ylabel('y');

    title(title_);
end



% parsing parameters, setting default values if necessary
% ###################################################### 
function task = parseParametersTwoDim(parameters)

    if nargin < 1
        parameters = {};
    else
        task = parameters;
    end

    if ~isfield(parameters, 'numTaskNeuronsPerConstraint')
        task.numTaskNeuronsPerConstraint = 10;
    end

    if ~isfield(parameters, 'numConstraints')
        task.numConstraints = 2;
    end
    if ~isfield(parameters, 'stateSpaceLimits')
        task.stateSpaceLimits = [-1, 1, -1, 1];
    end   
end
