% The MIT License (MIT)
% Copyright (c) 2016 Elmar Rueckert, Daniel Tanneberg, David Kappel
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE.

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: rueckert.elmar@gmail.com, tanneberg@ias.tu-darmstadt.de, david.kappel@igi.tugraz.at
% *
% **********************************************************
function net = GenerateSNNDiscreteNDS(W, b)

net.W = W;
net.b = b;

net.refractoryPeriodeInSeconds = 12e-3; %10ms
net.simTimeStep = 1e-3;
net.beta = 1; %inverse temperature

%The neurons’ spiking activities are defined by a stochastic model that assumes an
%exponential dependence between the membrane potential and the firing probability [Jolivet et al., 2006],
%which becomes the sigmoid function for discrete time steps, i.e.
%net.transferFunction = @(x) 1./(1. + exp(-x)); 

%The neural sampling networks are networks of spiking neurons
%that in their stationary stochastic dynamics perform MCMC sampling 
%from a particular probability distribution. The neuron used
%in these networks is a stochastic point neuron model with a firing
%probability density at time t equal to
net.transferFunction = @(x) 1/(net.refractoryPeriodeInSeconds/net.simTimeStep)*exp(x);

net.simulate = @(init_state_, numberOfTimeSteps_)localSimulate(init_state_, numberOfTimeSteps_, net);
%net.setNumberOfSpikesPerTimeStep = @(numOfSpikesPerStep_)localSetNumberOfSpikesPerTimeStep(numOfSpikesPerStep_, net);
net.PlotSpikeTrain = @(spikeTrain_)localPlotSpikeTrain(spikeTrain_, net);

end

function hfig = localPlotSpikeTrain(spikeTrain, net)
    
    hfig = figure;
    imagesc(spikeTrain);

end

% function net = localSetNumberOfSpikesPerTimeStep(numOfSpikesPerStep, net)
% 
%     %TODO compute additiveConstantlogW from simulating the net
%     additiveConstantlogW = 1;
%     net.W(net.W~=0) = net.W(net.W~=0) + additiveConstantlogW;
% end

function [spikeTrain, membranePot] = localSimulate(numberOfTimeSteps, net)

    %keyboard
    numNeurons = size(net.W,1);
    spikeTrain = sparse(numNeurons, numberOfTimeSteps);
    membranePot = sparse(numNeurons, numberOfTimeSteps);
    
    tau = net.refractoryPeriodeInSeconds/net.simTimeStep;
    refractoryState = zeros(numNeurons,1);

    for l=1:numberOfTimeSteps
        active = find(refractoryState > 0);
        inactive = find(refractoryState <= 0);
        
        %advance internal state of neurons that are in their ref. periode
        refractoryState(active) = max(zeros(length(active),1),-1+refractoryState(active));
        currSpikes = refractoryState > 0;
        %currSpikes = 5/tau*refractoryState;
        
        %update membran potential of all inactive neurons
        if l==1
            u = net.b(inactive,l);
        else
            u = net.W(:,inactive)'*currSpikes + net.b(inactive,l);
        end
        
        uSample = u;
        
        ableToSpikeIds = find(uSample ~= 0);
        s = rand(length(ableToSpikeIds),1) < net.transferFunction((net.beta)*(uSample(ableToSpikeIds)));
        
        if sum(find(net.transferFunction((net.beta)*(uSample(ableToSpikeIds))) > 1)) > 0
            %keyboard
        end
            
        %keyboard
        refractoryState(inactive(ableToSpikeIds),1) = tau*s;
        membranePot(inactive(ableToSpikeIds),l) = u(ableToSpikeIds);% > 0;
        %the neuron is active for tau time steps or cannot fire again
        %within the next tau steps?
        spikeTrain(inactive(ableToSpikeIds),l) = s;    
        %spikeTrain(inactive(ableToSpikeIds),l:l+tau-1) = repmat(s,1,tau);
        %keyboard
    end
    
end

function [f, spikes, c, refState] = NeuralDynamicsSampler(model, OPTIONS, f, W, b, beta, refState, tau)
    
    logTau = log(tau);
    inds = 1:model.n;
    
    c = 0;
    spikes = zeros(size(f));
    for i=1:model.n
        l=inds(i);
        active = find(refState(:,l) > 0);
        inactive = setdiff(1:size(f,1),active);
        %advance internal state of neurons that are in their ref. periode
        refState(active,l) = max(zeros(length(active),1),-1+refState(active,l));
        %update membran potential of all other neurons
        if l==1
            tmp = W(inactive,inactive)'*f(inactive,l+1);
            u = tmp + b(inactive,l);
%             figure(1); clf; stem(tmp); hold on; stem(b(inactive,l), 'r');
        elseif l==model.n
            tmp = W(inactive,inactive)'*f(inactive,l-1);
            u = tmp + b(inactive,l);
        else
            tmp = W(inactive,inactive)'*f(inactive,l-1) + W(inactive,inactive)'*f(inactive,l+1);
            u = tmp + b(inactive,l);
        end
        %spike
        if 1==0
            uSample = u - logTau;
%             u=u-max(u);
%             uSample = u ./ sum(u);
            s = rand(length(inactive),1) < compSigma((beta)*(uSample));
        else
           %WTA 
           if sum(active) == 0
               %u=u-max(u);
               u=exp(beta*u);
               if sum(isnan(u)) > 0
                   error('bad temperature values');
               end
               u=u./sum(u);
               dr=histc(rand(1,1),[0; cumsum(u)]);
               s = dr(1,1:end-1);
               if sum(s) < 1
                    keyboard
               end
           else
               s = zeros(size(f(inactive,l)));
           end
        end
        refState(inactive,l) = tau*s;
        f(inactive,l) = s;% > 0;
        spikes(inactive,l) = s;
        c = c + sum(tmp);
    end
    
end
