% The MIT License (MIT)
% Copyright (c) 2016  Daniel Tanneberg, David Kappel, Elmar Rueckert
% 
% Permission is hereby granted, free of charge, to any person obtaining a 
% copy of this software and associated documentation files (the "Software"), 
% to deal in the Software without restriction, including without limitation 
% the rights to use, copy, modify, merge, publish, distribute, sublicense, 
% and/or sell copies of the Software, and to permit persons to whom the 
% Software is furnished to do so, subject to the following conditions:
% 
% The above copyright notice and this permission notice shall be included
% in all copies or substantial portions of the Software.
% 
% THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
% OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
% FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
% THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
% LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
% FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
% DEALINGS IN THE SOFTWARE. 

% **********************************************************
% *
% * Supplementary Code to 
% *                       Recurrent Spiking Networks Solve Planning Tasks, 
% *                       Scientific Reports, 6, 21142, Nature Publishing Group.
% * 
% *                       Rueckert, Elmar and Kappel, David and Tanneberg, Daniel
% *                       and Pecevski, Dejan and Peters, Jan (2016).
% * 
% * If you use this code, please cite the paper!
% * 
% * contact: tanneberg@ias.tu-darmstadt.de, david.kappel@igi.tugraz.at, rueckert.elmar@gmail.com
% *
% **********************************************************
function [used_plans, unused_plans, ids] = rejectSamples(plans, task, delta, time_window, threshold)
% filter out samples that don't satisfy some constraints

    target = task.targetState;
    init = task.initialState;
    init_delay = task.numStepsInitialStateOn;
    
    ids = [];
    
    % filter only trajectories which reach the target & start at initial
    % 2D
    n = size(plans,1) / 2;
    used_plans = []; 
    unused_plans = [];
    init_dist = 0;
    target_dist = 0;
    
    gaussFilter = gausswin(time_window); %gaussian filter   
    gaussFilter = gaussFilter ./ sum(gaussFilter);
    
    for i=1:n
        t = (i*2)-1;
        x = plans(t,:);
        y = plans(t+1,:);
        
        dists = pdist2(init',[x(1:init_delay);y(1:init_delay)]');
        tmp = conv(dists', gaussFilter, 'same')';
        i_s = min(tmp);    
        
        init_dist = init_dist + i_s;
        
        dists = pdist2(target',[x(init_delay:end);y(init_delay:end)]');
        tmp = conv(dists', gaussFilter, 'same')';
        t_s = min(tmp);    
        
        target_dist = target_dist + t_s;

        if t_s <= delta  && i_s <= delta
           used_plans = [used_plans;x;y]; 
           ids = [ids,i];
        else
            unused_plans = [unused_plans;x;y];
        end
    end
    
    
    % filter trajs that are 'too fast'/too jerky
    n = size(used_plans,1) / 2;
    traj_len = size(used_plans,2);
    steps = 20;
    points = task.numStepsInitialStateOn:steps:traj_len-100;

    smoothness = zeros(1,n);
    for i=1:n
            t_i = (i*2)-1;
            x_i = used_plans(t_i,points);
            y_i = used_plans(t_i+1,points);
            smoothness(i) = evalTraj([x_i; y_i]);
    end
    
    accepted_ids = smoothness < threshold;
 
    new_plans = [];
    new_ids = [];
    for i=1:size(accepted_ids,2)
        if accepted_ids(i) == 1
            t_i = (i*2)-1;
            x_i = used_plans(t_i,:);
            y_i = used_plans(t_i+1,:);
            new_plans = [new_plans;x_i;y_i];
            new_ids = [new_ids,ids(i)];
        else
            t_i = (i*2)-1;
            x_i = used_plans(t_i,:);
            y_i = used_plans(t_i+1,:);
            unused_plans = [unused_plans;x_i;y_i];
        end
    end
    used_plans = new_plans;
    ids = ids(accepted_ids);
    

    % reject trajs that hit obstacle
    if task.isObstacle 
        new_plans = [];
        new_ids = [];
        for i=1:numel(ids)

            hit = 0;
            for o=1:numel(task.obstacle_ids)
                obst = task.neuronIdToStateMapping(:,task.obstacle_ids(o));
                t_i = (i*2)-1;
                x_i = used_plans(t_i,:);
                y_i = used_plans(t_i+1,:);

                dists = pdist2(obst',[x_i;y_i]');
                dists = dists(find(task.obstacle_times(o,:) == -10));
                if min(dists) < 0.1 % too close to obstacle
                    unused_plans = [unused_plans;x_i;y_i];
                    hit = 1;
                    break;
                end
            end
            if hit == 0
                new_ids = [new_ids,ids(i)];
                new_plans = [new_plans;x_i;y_i];
            end
        end
        used_plans = new_plans;
        ids = new_ids;
    end
end

